from aviary.variable_info.enums import SpeedType
from aviary.subsystems.propulsion.propulsion_builder import CorePropulsionBuilder
from aviary.subsystems.geometry.geometry_builder import CoreGeometryBuilder
from aviary.subsystems.mass.mass_builder import CoreMassBuilder
from aviary.subsystems.aerodynamics.aerodynamics_builder import CoreAerodynamicsBuilder
from aviary.variable_info.variable_meta_data import _MetaData as BaseMetaData, Mission
from aviary.variable_info.enums import LegacyCode


GASP = LegacyCode.GASP

prop = CorePropulsionBuilder('core_propulsion', BaseMetaData)
mass = CoreMassBuilder('core_mass', BaseMetaData, GASP)
aero = CoreAerodynamicsBuilder('core_aerodynamics', BaseMetaData, GASP)
geom = CoreGeometryBuilder('core_geometry', BaseMetaData, GASP)

default_premission_subsystems = [prop, geom, aero, mass]
default_mission_subsystems = [aero, prop]

mission_distance = 3675

phase_info = {
    'groundroll': {
        'user_options': {
            'num_segments': 1,
            'order': 3,
            'connect_initial_mass': False,
            'fix_initial': True,
            'fix_initial_mass': False,
            'duration_ref': (50., 's'),
            'duration_bounds': ((1., 100.), 's'),
            'velocity_lower': (0, 'kn'),
            'velocity_upper': (1000, 'kn'),
            'velocity_ref': (150, 'kn'),
            'mass_lower': (0, 'lbm'),
            'mass_upper': (None, 'lbm'),
            'mass_ref': (150_000, 'lbm'),
            'mass_defect_ref': (150_000, 'lbm'),
            'distance_lower': (0, 'ft'),
            'distance_upper': (10.e3, 'ft'),
            'distance_ref': (3000, 'ft'),
            'distance_defect_ref': (3000, 'ft'),
        },
        'initial_guesses': {
            'times': ([0.0, 40.0], 's'),
            'velocity': ([0.066, 143.1], 'kn'),
            'distance': ([0.0, 1000.], 'ft'),
            'throttle': ([0.956, 0.956], 'unitless'),
        }
    },
    'rotation': {
        'user_options': {
            'num_segments': 1,
            'order': 3,
            'fix_initial': False,
            'duration_bounds': ((1, 100), 's'),
            'duration_ref': (50.0, 's'),
            'velocity_lower': (0, 'kn'),
            'velocity_upper': (1000, 'kn'),
            'velocity_ref': (100, 'kn'),
            'velocity_ref0': (0, 'kn'),
            'mass_lower': (0, 'lbm'),
            'mass_upper': (None, 'lbm'),
            'mass_ref': (150_000, 'lbm'),
            'mass_defect_ref': (150_000, 'lbm'),
            'distance_lower': (0, 'ft'),
            'distance_upper': (10_000, 'ft'),
            'distance_ref': (5000, 'ft'),
            'distance_defect_ref': (5000, 'ft'),
            'angle_lower': (0., 'rad'),
            'angle_upper': (5., 'rad'),
            'angle_ref': (5., 'rad'),
            'angle_defect_ref': (5., 'rad'),
            'normal_ref': (10000, 'lbf'),
        },
        'initial_guesses': {
            'times': ([40.0, 5.0], 's'),
            'alpha': ([0.0, 2.5], 'deg'),
            'velocity': ([143, 150.], 'kn'),
            'distance': ([3680.37217765, 4000], 'ft'),
            'throttle': ([0.956, 0.956], 'unitless'),
        }
    },
    'ascent': {
        'user_options': {
            'num_segments': 4,
            'order': 3,
            'fix_initial': False,
            'velocity_lower': (0, 'kn'),
            'velocity_upper': (700, 'kn'),
            'velocity_ref': (200, 'kn'),
            'velocity_ref0': (0, 'kn'),
            'mass_lower': (0, 'lbm'),
            'mass_upper': (None, 'lbm'),
            'mass_ref': (150_000, 'lbm'),
            'mass_defect_ref': (150_000, 'lbm'),
            'distance_lower': (0, 'ft'),
            'distance_upper': (15_000, 'ft'),
            'distance_ref': (1e4, 'ft'),
            'distance_defect_ref': (1e4, 'ft'),
            'alt_lower': (0, 'ft'),
            'alt_upper': (700, 'ft'),
            'alt_ref': (1000, 'ft'),
            'alt_defect_ref': (1000, 'ft'),
            'final_altitude': (500, 'ft'),
            'alt_constraint_ref': (500, 'ft'),
            'angle_lower': (-10, 'rad'),
            'angle_upper': (20, 'rad'),
            'angle_ref': (57.2958, 'deg'),
            'angle_defect_ref': (57.2958, 'deg'),
            'pitch_constraint_lower': (0., 'deg'),
            'pitch_constraint_upper': (15., 'deg'),
            'pitch_constraint_ref': (1., 'deg'),
        },
        'initial_guesses': {
            'times': ([45., 25.], 's'),
            'flight_path_angle': ([0.0, 8.], 'deg'),
            'alpha': ([2.5, 1.5], 'deg'),
            'velocity': ([150., 185.], 'kn'),
            'distance': ([4.e3, 10.e3], 'ft'),
            'altitude': ([0.0, 500.], 'ft'),
            'tau_gear': (0.2, 'unitless'),
            'tau_flaps': (0.9, 'unitless'),
            'throttle': ([0.956, 0.956], 'unitless'),
        }
    },
    'accel': {
        'user_options': {
            'num_segments': 1,
            'order': 3,
            'fix_initial': False,
            'alt': (500, 'ft'),
            'EAS_constraint_eq': (250, 'kn'),
            'duration_bounds': ((1, 200), 's'),
            'duration_ref': (1000, 's'),
            'velocity_lower': (150, 'kn'),
            'velocity_upper': (270, 'kn'),
            'velocity_ref': (250, 'kn'),
            'velocity_ref0': (150, 'kn'),
            'mass_lower': (0, 'lbm'),
            'mass_upper': (None, 'lbm'),
            'mass_ref': (150_000, 'lbm'),
            'mass_defect_ref': (150_000, 'lbm'),
            'distance_lower': (0, 'NM'),
            'distance_upper': (150, 'NM'),
            'distance_ref': (5, 'NM'),
            'distance_defect_ref': (5, 'NM'),
        },
        'initial_guesses': {
            'times': ([70., 13.], 's'),
            'velocity': ([185., 250.], 'kn'),
            'distance': ([10.e3, 20.e3], 'ft'),
            'throttle': ([0.956, 0.956], 'unitless'),
        }
    },
    'climb1': {
        'user_options': {
            'num_segments': 1,
            'order': 3,
            'fix_initial': False,
            'EAS_target': (250, 'kn'),
            'mach_cruise': 0.8,
            'target_mach': False,
            'final_altitude': (10.e3, 'ft'),
            'duration_bounds': ((30, 300), 's'),
            'duration_ref': (1000, 's'),
            'alt_lower': (400, 'ft'),
            'alt_upper': (11_000, 'ft'),
            'alt_ref': (10.e3, 'ft'),
            'mass_lower': (0, 'lbm'),
            'mass_upper': (None, 'lbm'),
            'mass_ref': (150_000, 'lbm'),
            'mass_defect_ref': (150_000, 'lbm'),
            'distance_lower': (0, 'NM'),
            'distance_upper': (500, 'NM'),
            'distance_ref': (10, 'NM'),
            'distance_ref0': (0, 'NM'),
        },
        'initial_guesses': {
            'times': ([1., 2.], 'min'),
            'distance': ([20.e3, 100.e3], 'ft'),
            'altitude': ([500., 10.e3], 'ft'),
            'throttle': ([0.956, 0.956], 'unitless'),
        }
    },
    'climb2': {
        'user_options': {
            'num_segments': 3,
            'order': 3,
            'fix_initial': False,
            'EAS_target': (270, 'kn'),
            'mach_cruise': 0.8,
            'target_mach': True,
            'final_altitude': (37.5e3, 'ft'),
            'required_available_climb_rate': (0.1, 'ft/min'),
            'duration_bounds': ((200, 17_000), 's'),
            'duration_ref': (5000, 's'),
            'alt_lower': (9000, 'ft'),
            'alt_upper': (40000, 'ft'),
            'alt_ref': (30000, 'ft'),
            'alt_ref0': (0, 'ft'),
            'mass_lower': (0, 'lbm'),
            'mass_upper': (None, 'lbm'),
            'mass_ref': (150_000, 'lbm'),
            'mass_defect_ref': (150_000, 'lbm'),
            'distance_lower': (10, 'NM'),
            'distance_upper': (1000, 'NM'),
            'distance_ref': (500, 'NM'),
            'distance_ref0': (0, 'NM'),
            'distance_defect_ref': (500, 'NM'),
        },
        'initial_guesses': {
            'times': ([216., 1300.], 's'),
            'distance': ([100.e3, 200.e3], 'ft'),
            'altitude': ([10.e3, 37.5e3], 'ft'),
            'throttle': ([0.956, 0.956], 'unitless'),
        }
    },
    'cruise': {
        'user_options': {
            'alt_cruise': (37.5e3, 'ft'),
            'mach_cruise': 0.8,
        },
        'initial_guesses': {
            # [Initial mass, delta mass] for special cruise phase.
            'mass': ([171481., -35000], 'lbm'),
            'initial_distance': (200.e3, 'ft'),
            'initial_time': (1516., 's'),
            'altitude': (37.5e3, 'ft'),
            'mach': (0.8, 'unitless'),
        }
    },
    'desc1': {
        'user_options': {
            'num_segments': 3,
            'order': 3,
            'fix_initial': False,
            'input_initial': False,
            'EAS_limit': (350, 'kn'),
            'mach_cruise': 0.8,
            'input_speed_type': SpeedType.MACH,
            'final_altitude': (10.e3, 'ft'),
            'duration_bounds': ((300., 900.), 's'),
            'duration_ref': (1000, 's'),
            'alt_lower': (1000, 'ft'),
            'alt_upper': (40_000, 'ft'),
            'alt_ref': (30_000, 'ft'),
            'alt_ref0': (0, 'ft'),
            'alt_constraint_ref': (10000, 'ft'),
            'mass_lower': (0, 'lbm'),
            'mass_upper': (None, 'lbm'),
            'mass_ref': (140_000, 'lbm'),
            'mass_ref0': (0, 'lbm'),
            'mass_defect_ref': (140_000, 'lbm'),
            'distance_lower': (3000., 'NM'),
            'distance_upper': (5000., 'NM'),
            'distance_ref': (mission_distance, 'NM'),
            'distance_ref0': (0, 'NM'),
            'distance_defect_ref': (100, 'NM'),
        },
        'initial_guesses': {
            'mass': (136000., 'lbm'),
            'altitude': ([37.5e3, 10.e3], 'ft'),
            'throttle': ([0.0, 0.0], 'unitless'),
            'distance': ([.92*mission_distance, .96*mission_distance], 'NM'),
            'times': ([28000., 500.], 's'),
        }
    },
    'desc2': {
        'user_options': {
            'num_segments': 1,
            'order': 7,
            'fix_initial': False,
            'input_initial': False,
            'EAS_limit': (250, 'kn'),
            'mach_cruise': 0.80,
            'input_speed_type': SpeedType.EAS,
            'final_altitude': (1000, 'ft'),
            'duration_bounds': ((100., 5000), 's'),
            'duration_ref': (500, 's'),
            'alt_lower': (500, 'ft'),
            'alt_upper': (11_000, 'ft'),
            'alt_ref': (10.e3, 'ft'),
            'alt_ref0': (1000, 'ft'),
            'alt_constraint_ref': (1000, 'ft'),
            'mass_lower': (0, 'lbm'),
            'mass_upper': (None, 'lbm'),
            'mass_ref': (150_000, 'lbm'),
            'mass_defect_ref': (150_000, 'lbm'),
            'distance_lower': (0, 'NM'),
            'distance_upper': (5000, 'NM'),
            'distance_ref': (3500, 'NM'),
            'distance_defect_ref': (100, 'NM'),
        },
        'initial_guesses': {
            'mass': (136000., 'lbm'),
            'altitude': ([10.e3, 1.e3], 'ft'),
            'throttle': ([0., 0.], 'unitless'),
            'distance': ([.96*mission_distance, mission_distance], 'NM'),
            'times': ([28500., 500.], 's'),
        }
    },
}


def phase_info_parameterization(phase_info, post_mission_info, aviary_inputs):
    """
    Modify the values in the phase_info dictionary to accomodate different values
    for the following mission design inputs: cruise altitude, cruise mach number,
    cruise range, design gross mass.

    Parameters
    ----------
    phase_info : dict
        Dictionary of phase settings for a mission profile
    aviary_inputs : <AviaryValues>
        Object containing values and units for all aviary inputs and options

    Returns
    -------
    dict
        Modified phase_info that has been changed to match the new mission
        parameters
    """

    range_cruise = aviary_inputs.get_val(Mission.Design.RANGE, units='NM')
    alt_cruise = aviary_inputs.get_val(Mission.Design.CRUISE_ALTITUDE, units='ft')
    gross_mass = aviary_inputs.get_val(Mission.Design.GROSS_MASS, units='lbm')
    mach_cruise = aviary_inputs.get_val(Mission.Design.MACH)

    # Range
    old_range_cruise = phase_info['desc2']['initial_guesses']['distance'][0][1]
    range_scale = 1.0
    if range_cruise != old_range_cruise:

        phase_info['desc1']['initial_guesses']['distance'] = \
            ([.92*range_cruise, .96*range_cruise], 'NM')
        phase_info['desc2']['initial_guesses']['distance'] = \
            ([.96*range_cruise, range_cruise], 'NM')
        range_scale = range_cruise / old_range_cruise

    # Altitude
    old_alt_cruise = phase_info['climb2']['user_options']['final_altitude'][0]
    if alt_cruise != old_alt_cruise:

        phase_info['climb2']['user_options']['final_altitude'] = (alt_cruise, 'ft')
        phase_info['climb2']['initial_guesses']['altitude'] = ([10.e3, alt_cruise], 'ft')
        phase_info['cruise']['initial_guesses']['altitude'] = (alt_cruise, 'ft')
        phase_info['desc1']['initial_guesses']['altitude'] = ([alt_cruise, 10.e3], 'ft')

        # TODO - Could adjust time guesses/bounds in climb2 and desc2.

    # Mass
    old_gross_mass = 175400.0
    if gross_mass != old_gross_mass:

        # Note, this requires that the guess for gross mass is pretty close to the
        # compute mass.

        fuel_used = 35000 * range_scale
        phase_info['groundroll']['initial_guesses']['mass'] = \
            ([gross_mass, gross_mass], 'lbm')
        phase_info['rotation']['initial_guesses']['mass'] = \
            ([gross_mass, gross_mass], 'lbm')
        phase_info['accel']['initial_guesses']['mass'] = \
            ([gross_mass, gross_mass], 'lbm')
        phase_info['ascent']['initial_guesses']['mass'] = \
            ([gross_mass, gross_mass], 'lbm')

        phase_info['cruise']['initial_guesses']['mass'] = \
            ([gross_mass, -fuel_used], 'lbm')

        end_mass = gross_mass - fuel_used
        phase_info['desc1']['initial_guesses']['mass'] = (end_mass, 'lbm')
        phase_info['desc2']['initial_guesses']['mass'] = (end_mass, 'lbm')

    # Mach
    old_mach_cruise = phase_info['cruise']['initial_guesses']['mach'][0]
    if mach_cruise != old_mach_cruise:

        phase_info['cruise']['initial_guesses']['mach'] = (mach_cruise, 'unitless')

    return phase_info, post_mission_info



from openmdao.utils.units import valid_units
from aviary.variable_info.enums import SpeedType, EquationsOfMotion

TWO_DEGREES_OF_FREEDOM = EquationsOfMotion.TWO_DEGREES_OF_FREEDOM
HEIGHT_ENERGY = EquationsOfMotion.HEIGHT_ENERGY
SOLVED = EquationsOfMotion.SOLVED


def check_phase_info(phase_info, mission_method):
    # Define common keys for all phases
    common_keys = {
        'num_segments': int,
        'order': int,
        'fix_initial': (bool, dict),
    }

    # Common key-values for climb, cruise, and descent
    common_entries = {
        'optimize_mach': bool,
        'optimize_altitude': bool,
        'solve_for_distance': bool,
        'initial_mach': tuple,
        'final_mach': tuple,
        'mach_bounds': tuple,
        'initial_altitude': tuple,
        'final_altitude': tuple,
        'altitude_bounds': tuple,
        'throttle_enforcement': str,
        'constrain_final': bool,
        'fix_duration': bool,
        'initial_bounds': tuple,
        'duration_bounds': tuple,
    }

    # Combine common and phase-specific entries
    phase_keys_height_energy = {
        'pre_mission': {'include_takeoff': bool, 'optimize_mass': bool},
        'post_mission': {'include_landing': bool}
    }

    common_TAS = {
        'velocity_lower': tuple,
        'velocity_upper': tuple,
        'velocity_ref': tuple,
    }
    common_mass = {
        'mass_lower': tuple,
        'mass_upper': tuple,
        'mass_ref': tuple,
        'mass_defect_ref': tuple,
    }
    common_distance = {
        'distance_lower': tuple,
        'distance_upper': tuple,
        'distance_ref': tuple,
    }
    common_angle = {
        'angle_lower': tuple,
        'angle_upper': tuple,
        'angle_ref': tuple,
        'angle_defect_ref': tuple,
    }
    common_duration = {
        'duration_bounds': tuple,
        'duration_ref': tuple,
    }
    common_alt = {
        'alt_lower': tuple,
        'alt_upper': tuple,
        'alt_ref': tuple,
    }
    common_descent = {
        'input_initial': (bool, dict),
        'EAS_limit': tuple,
        'mach_cruise': float,
        'input_speed_type': SpeedType,
        'final_altitude': tuple,
        'alt_constraint_ref': tuple,
    }

    phase_keys_gasp = {
        'groundroll': {
            'connect_initial_mass': bool,
            'fix_initial_mass': bool,
            **common_duration,
            **common_TAS,
            **common_mass,
            **common_distance,
            'distance_defect_ref': tuple,
        },
        'rotation': {
            **common_duration,
            **common_TAS,
            **common_mass,
            **common_distance,
            **common_angle,
            'normal_ref': tuple,
            'velocity_ref0': tuple,
            'distance_defect_ref': tuple,
        },
        'ascent': {
            **common_TAS,
            **common_mass,
            **common_distance,
            **common_alt,
            'final_altitude': tuple,
            'alt_constraint_ref': tuple,
            'alt_defect_ref': tuple,
            **common_angle,
            'pitch_constraint_lower': tuple,
            'pitch_constraint_upper': tuple,
            'pitch_constraint_ref': tuple,
            'velocity_ref0': tuple,
            'distance_defect_ref': tuple,
        },
        'accel': {
            'alt': tuple,
            'EAS_constraint_eq': tuple,
            **common_duration,
            'duration_ref': tuple,
            **common_TAS,
            **common_mass,
            **common_distance,
            'velocity_ref0': tuple,
            'distance_defect_ref': tuple,
        },
        'climb1': {
            'EAS_target': tuple,
            'mach_cruise': float,
            'target_mach': bool,
            'final_altitude': tuple,
            **common_duration,
            **common_alt,
            **common_mass,
            **common_distance,
            'distance_ref0': tuple,
        },
        'climb2': {
            'EAS_target': tuple,
            'mach_cruise': float,
            'target_mach': bool,
            'final_altitude': tuple,
            'required_available_climb_rate': tuple,
            **common_duration,
            **common_alt,
            **common_mass,
            **common_distance,
            'alt_ref0': tuple,
            'distance_ref0': tuple,
            'distance_defect_ref': tuple,
        },
        'cruise': {
            'mach_cruise': float,
            'alt_cruise': tuple,
        },
        'desc1': {
            **common_descent,
            **common_duration,
            **common_alt,
            **common_mass,
            **common_distance,
            'alt_ref0': tuple,
            'mass_ref0': tuple,
            'distance_ref0': tuple,
        },
        'desc2': {
            **common_descent,
            **common_duration,
            **common_alt,
            **common_mass,
            **common_distance,
            'alt_ref0': tuple,
            'distance_defect_ref': tuple,
        },
    }

    phase_keys = {}
    if mission_method is TWO_DEGREES_OF_FREEDOM:
        for phase in phase_info:
            if phase != 'pre_mission' and phase != 'post_mission':
                if phase == 'cruise':
                    phase_keys[phase] = {**phase_keys_gasp[phase]}
                else:
                    phase_keys[phase] = {**common_keys, **phase_keys_gasp[phase]}
    elif mission_method is SOLVED:
        return
    elif mission_method is HEIGHT_ENERGY:
        for phase in phase_info:
            if phase != 'pre_mission' and phase != 'post_mission':
                phase_keys[phase] = {**common_keys, **common_entries}
            else:
                phase_keys[phase] = phase_keys_height_energy[phase]
    else:
        possible_values = ["'"+e.value+"'" for e in EquationsOfMotion]
        possible_values[-1] = "or " + possible_values[-1]
        raise ValueError("Invalid mission_method. Please choose from " +
                         ", ".join(possible_values) + ".")

    # Check if all phases exist in phase_info
    for phase in phase_info:
        if 'user_options' in phase_info[phase]:
            phase_options = phase_info[phase]['user_options']
        else:
            phase_options = phase_info[phase]

        # Check if all required keys exist, if there are no extra keys, and if they are of the correct type
        for key, expected_type in phase_keys[phase].items():
            # Check tuples for (val, units) structure
            if expected_type is tuple:
                if key not in phase_options:
                    raise ValueError(
                        f"Key {key} not found in phase {phase} of the provided dictionary.")
                value = phase_options[key]
                if not isinstance(value, tuple) or len(value) != 2:
                    raise ValueError(f"Key {key} in phase {phase} should be a tuple of size 2 (value, unit), "
                                     f"but got {value}.")
                unit = value[1]
                if not valid_units(unit):
                    raise ValueError(
                        f"Key {key} in phase {phase} has an invalid unit {unit}.")

            if not isinstance(phase_options[key], expected_type):
                raise TypeError(f"Key {key} in phase {phase} should be of type {expected_type.__name__}, "
                                f"but got type {type(phase_options[key]).__name__}")

    return True



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.utils.aviary_values import AviaryValues
from aviary.mission.gasp_based.phases.taxi_component import TaxiFuelComponent
from aviary.variable_info.variables import Dynamic, Mission


class TaxiFuelComponentTestCase(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem(model=om.Group())

        aviary_options = AviaryValues()
        aviary_options.set_val(Mission.Taxi.DURATION, 0.1677, units="h")

        self.prob.model.add_subsystem('taxi', TaxiFuelComponent(
            aviary_options=aviary_options), promotes=['*'])

    def test_fuel_consumed(self):
        self.prob.setup(force_alloc_complex=True)

        self.prob.set_val(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, -1512, units="lbm/h")

        self.prob.run_model()

        assert_near_equal(self.prob["taxi_fuel_consumed"], 1512 * 0.1677, 1e-6)
        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, get_units
from aviary.variable_info.variables import Aircraft


class SkinFrictionDrag(om.ExplicitComponent):
    """
    Computes the total coefficient of drag due to skin friction from the skin friction
    coefficients of each component surface.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Form factor fit coefficients.
        self.F = np.array([
            4.34255, -1.14281, .171203, -.0138334, .621712e-3, .137442e-6, -.145532e-4,
            2.94206, 7.16974, 48.8876, -1403.02, 8598.76, -15834.3, 4.275])

    def initialize(self):
        """
        Declare options.
        """
        self.options.declare(
            'num_nodes', types=int, default=1,
            desc='The number of points at which the cross product is computed.')
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

        # TODO: Convert these into aviary_options entries.
        self.options.declare(
            'excrescences_drag', default=0.06,
            desc='Drag contribution of excrescences as a percentage.')

    def setup(self):
        aviary_options: AviaryValues = self.options['aviary_options']
        nn = self.options['num_nodes']

        zero_count = (0, None)
        nvtail, _ = aviary_options.get_item(Aircraft.VerticalTail.NUM_TAILS, zero_count)
        nfuse, _ = aviary_options.get_item(Aircraft.Fuselage.NUM_FUSELAGES, zero_count)
        num_engines, _ = aviary_options.get_item(Aircraft.Engine.NUM_ENGINES, zero_count)
        self.nc = nc = 2 + nvtail + nfuse + int(sum(num_engines))

        # Computed by other components in drag group.
        self.add_input('skin_friction_coeff', np.zeros((nn, nc)), units='unitless')
        self.add_input('Re', np.ones((nn, nc)), units='unitless')

        # These have been assembled from the individually-titled component variables.
        self.add_input(
            'fineness_ratios', np.ones(nc),
            desc='Vector of component fineness ratios.',
            units='unitless')
        self.add_input(
            'wetted_areas', np.ones(nc), units=get_units(Aircraft.Wing.AREA),
            desc='Vector of component wetted areas.')
        self.add_input(
            'laminar_fractions_upper', np.ones(nc), units=get_units(Aircraft.Wing.LAMINAR_FLOW_UPPER),
            desc='Vector of component upper-surface laminar-flow fractions.')
        self.add_input(
            'laminar_fractions_lower', np.ones(nc), units=get_units(Aircraft.Wing.LAMINAR_FLOW_LOWER),
            desc='Vector of component lower-surface laminar-flow fractions.')

        # Aircraft design inputs
        add_aviary_input(self, Aircraft.Wing.AREA, 0.0)

        # Output
        self.add_output(
            'skin_friction_drag_coeff', np.zeros(nn), units='unitless',
            desc='Skin friction drag coefficient.')

    def setup_partials(self):
        nn = self.options["num_nodes"]
        nc = self.nc
        n = nn * nc

        self.declare_partials(
            of='skin_friction_drag_coeff',
            wrt=[Aircraft.Wing.AREA])

        rows = np.repeat(np.arange(nn), nc)
        cols = np.tile(np.arange(nc), nn)
        self.declare_partials(
            of='skin_friction_drag_coeff',
            wrt=['fineness_ratios', 'wetted_areas',
                 'laminar_fractions_upper', 'laminar_fractions_lower'], rows=rows, cols=cols)

        cols = np.arange(n)
        self.declare_partials(
            of='skin_friction_drag_coeff',
            wrt=['skin_friction_coeff', 'Re'], rows=rows, cols=cols)

    def compute(self, inputs, outputs):
        nc = self.nc
        aviary_options: AviaryValues = self.options['aviary_options']

        cf = inputs['skin_friction_coeff']
        Re = inputs['Re']

        fineness = inputs['fineness_ratios']
        wetted_area = inputs['wetted_areas']
        lam_up = inputs['laminar_fractions_upper']
        lam_low = inputs['laminar_fractions_lower']
        mission_wing_area = inputs[Aircraft.Wing.AREA]

        laminar_flow = np.any(lam_up > 0.0) or np.any(lam_low > 0.0)

        if laminar_flow:
            laminar_upper = _calc_laminar_flow(lam_up)
            laminar_lower = _calc_laminar_flow(lam_low)
            cf = cf - 0.5 * (cf - 1.328 / np.sqrt(Re)) * \
                (laminar_lower + laminar_upper)

        form_factor = np.empty(nc, dtype=cf.dtype)

        F = self.F

        # Form factor for bodies.
        idx_body = np.where(fineness > 0.5)[0]
        fine = fineness[idx_body]

        # Note: this equation is implemented exactly as it is in FLOPS. Terms 5 and 6 in the
        # Horner expansion seem to be out of order (cf. F[5] + fine * F[6]), and the origin
        # of this equation is not clear.
        # However, if you swap the terms, you end up with negative skin friction coef.
        form_factor[idx_body] = F[0] + fine * \
            (F[1] + fine * (F[2] + fine * (F[3] + fine * (F[4] + fine * (F[5] * fine + F[6])))))

        idx_max = np.where(fineness >= 20.0)
        form_factor[idx_max] = 1.0

        # Form factor for surfaces.
        idx_surf = np.where(fineness <= 0.5)[0]
        fine = fineness[idx_surf]
        airfoil = aviary_options.get_val(Aircraft.Wing.AIRFOIL_TECHNOLOGY)

        FF1 = 1.0 + fine * (F[7] + fine * (F[8] + fine *
                            (F[9] + fine * (F[10] + fine * (F[11] + fine * F[12])))))
        FF2 = 1.0 + fine * self.F[13]

        form_factor[idx_surf] = FF1 * (2.0 - airfoil) + FF2 * (airfoil - 1.0)

        CDF = np.einsum('j,ij,j->i', wetted_area, cf, form_factor) * \
            (1.0 / mission_wing_area)

        # Add drag for excrescences.

        # TODO - Per component not completely implemented in aviary 1.0
        # Var mission_skin_friction_drag_corrections_count is a vector over components and is added
        # to the drag.
        # This may be "dead weight" from FLOPS - D.J.

        # An additional six percent of the skin friction drag is added to for excrescences
        # (miscellaneous).
        CDF *= 1.0 + self.options['excrescences_drag']

        outputs['skin_friction_drag_coeff'] = CDF

    def compute_partials(self, inputs, partials):
        nc = self.nc
        nn = self.options["num_nodes"]
        aviary_options: AviaryValues = self.options['aviary_options']

        cf = inputs['skin_friction_coeff']
        Re = inputs['Re']

        fineness = inputs['fineness_ratios']
        wetted_area = inputs['wetted_areas']
        lam_up = inputs['laminar_fractions_upper']
        lam_low = inputs['laminar_fractions_lower']
        mission_wing_area = inputs[Aircraft.Wing.AREA]

        laminar_flow = np.any(lam_up > 0.0) or np.any(lam_low > 0.0)

        den = 1.0 / np.sqrt(Re)
        lam_lam = -0.5 * (cf - 1.328 * den)
        if laminar_flow:
            laminar_upper = _calc_laminar_flow(lam_up)
            laminar_lower = _calc_laminar_flow(lam_low)
            lam_sum = laminar_lower + laminar_upper
            lam_cf = 1.0 - 0.5 * lam_sum
            lam_Re = -0.25 * 1.328 / Re ** 1.5 * lam_sum

            cf = cf - 0.5 * (cf - 1.328 * den) * lam_sum

        form_factor = np.empty(nc, dtype=cf.dtype)
        dform_dfine = np.empty(nc, dtype=cf.dtype)

        F = self.F

        # Form factor for bodies.
        idx_body = np.where(fineness > 0.5)[0]
        fine = fineness[idx_body]
        form_factor[idx_body] = F[0] + fine * \
            (F[1] + fine * (F[2] + fine * (F[3] + fine * (F[4] + fine * (F[5] * fine + F[6])))))
        dform_dfine[idx_body] = F[1] + fine * (2.0 * F[2] + fine * (
            3.0 * F[3] + fine * (4.0 * F[4] + fine * (6.0 * F[5] * fine + 5.0 * F[6]))))

        # When pinned above max fineness, deriv is zero.
        idx_max = np.where(fineness >= 20.0)
        form_factor[idx_max] = 1.0
        dform_dfine[idx_max] = 0.0

        # Form factor for surfaces.
        idx_surf = np.where(fineness <= 0.5)[0]
        fine = fineness[idx_surf]

        airfoil = aviary_options.get_val(Aircraft.Wing.AIRFOIL_TECHNOLOGY)

        FF1 = 1.0 + fine * (
            F[7] + fine
            * (F[8] + fine * (F[9] + fine * (F[10] + fine * (F[11] + fine * F[12])))))
        FF2 = 1.0 + fine * self.F[13]
        dFF1 = F[7] + fine * (
            2.0 * F[8] + fine * (
                3.0 * F[9] + fine
                * (4.0 * F[10] + fine * (5.0 * F[11] + fine * 6.0 * F[12]))))
        dFF2 = self.F[13]

        form_factor[idx_surf] = FF1 * (2.0 - airfoil) + FF2 * (airfoil - 1.0)
        dform_dfine[idx_surf] = dFF1 * (2.0 - airfoil) + dFF2 * (airfoil - 1.0)

        den = 1.0 / mission_wing_area
        CDF = np.einsum('j,ij,j->i', wetted_area, cf, form_factor) * den

        excr = 1.0 + self.options['excrescences_drag']
        CDF *= excr
        DCDF_dwet = excr * np.einsum('ij,j->ij', cf, form_factor) * den
        DCDF_dcf = excr * wetted_area * form_factor * den
        DCDF_dform = excr * np.einsum('j,ij->ij', wetted_area, cf) * den
        DCDF_dmwa = -excr * np.einsum('j,ij,j->i', wetted_area,
                                      cf, form_factor) * den ** 2

        DCDF_dlamup = DCDF_dcf * lam_lam * _calc_laminar_flow_deriv(lam_up)
        DCDF_dlamlow = DCDF_dcf * lam_lam * _calc_laminar_flow_deriv(lam_low)

        if laminar_flow:
            DCDF_dRe = np.einsum('j,ij->ij', DCDF_dcf, lam_Re)
            DCDF_dcf *= lam_cf

        partials['skin_friction_drag_coeff', 'skin_friction_coeff'] = np.tile(
            DCDF_dcf, nn)

        if laminar_flow:
            partials['skin_friction_drag_coeff', 'Re'] = DCDF_dRe.ravel()

        partials['skin_friction_drag_coeff', 'fineness_ratios'] = np.einsum(
            'ij,j->ij', DCDF_dform, dform_dfine).ravel()

        partials['skin_friction_drag_coeff', 'wetted_areas'] = DCDF_dwet.ravel()

        partials['skin_friction_drag_coeff',
                 'laminar_fractions_upper'] = DCDF_dlamup.ravel()

        partials['skin_friction_drag_coeff',
                 'laminar_fractions_lower'] = DCDF_dlamlow.ravel()

        partials['skin_friction_drag_coeff', Aircraft.Wing.AREA] = DCDF_dmwa.ravel()


def _calc_laminar_flow(lam):
    return lam * (0.0064164 + lam * (0.48087e-4 - 0.12234e-6 * lam))


def _calc_laminar_flow_deriv(lam):
    return 0.0064164 + lam * (0.96174e-4 - 0.36702e-6 * lam)



import openmdao.api as om

from aviary.subsystems.geometry.gasp_based.electric import CableSize
from aviary.subsystems.geometry.gasp_based.empennage import EmpennageSize
from aviary.subsystems.geometry.gasp_based.engine import EngineSize
from aviary.subsystems.geometry.gasp_based.fuselage import FuselageGroup
from aviary.subsystems.geometry.gasp_based.wing import WingGroup
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft


class SizeGroup(om.Group):

    """
    Group to pull together all the different components and subgroups of the SIZE subroutine

    """

    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):
        aviary_options = self.options['aviary_options']

        self.add_subsystem(
            "fuselage",
            FuselageGroup(
                aviary_options=aviary_options,
            ),
            promotes_inputs=["aircraft:*"],
            promotes_outputs=["aircraft:*"],
        )

        self.add_subsystem(
            "wing",
            WingGroup(
                aviary_options=aviary_options,
            ),
            promotes=["aircraft:*", "mission:*"],
        )

        self.add_subsystem(
            "empennage",
            EmpennageSize(aviary_options=aviary_options,),
            promotes=["aircraft:*"],
        )

        self.add_subsystem(
            "engine",
            EngineSize(aviary_options=aviary_options,),
            promotes_inputs=["aircraft:*"],
            promotes_outputs=["aircraft:*"],
        )

        if self.options["aviary_options"].get_val(Aircraft.Electrical.HAS_HYBRID_SYSTEM, units='unitless'):
            self.add_subsystem(
                "cable",
                CableSize(aviary_options=aviary_options,),
                promotes_inputs=["aircraft:*"],
                promotes_outputs=["aircraft:*"],
            )

        self.set_input_defaults(Aircraft.Fuselage.AVG_DIAMETER, units="inch")



'''
Define utilities for building phases.

Classes
-------
InitialGuess : a utility for setting an initial guess on a problem

InitialGuessControl : a utility for setting an initial guess for a control on a problem

InitialGuessParameter : a utility for setting an initial guess for a parameter on a
problem

InitialGuessPolynomialControl : a utility for setting an initial guess for a polynomial
control on a problem

InitialGuessState : a utility for setting an initial guess for a state on a problem

InitialGuessTime : a utility for setting guesses for initial time and duration on a
problem
'''
from collections.abc import Sequence

import dymos as dm
import numpy as np
import openmdao.api as om


class InitialGuess:
    '''
    Define a utility for setting an initial guess on a problem.

    Attributes
    ----------
    key : str
        base name for the initial guess
    '''
    __slots__ = ('key',)

    def __init__(self, key):
        self.key = key

    def apply_initial_guess(
        self, prob: om.Problem, traj_name, phase: dm.Phase, phase_name, val, units
    ):
        '''
        Set the initial guess on the problem.
        '''
        complete_key = self._get_complete_key(traj_name, phase_name)

        # TODO: this is a short term hack in need of an appropriate long term solution
        #    - to interpolate, or not to interpolate: that is the question
        #    - the solution should probably be a value decoration (a wrapper) that is
        #      both lightweight and easy to check and unpack
        if (
            isinstance(val, np.ndarray)
            or (isinstance(val, Sequence) and not isinstance(val, str))
        ):
            val = phase.interp(self.key, val)

        prob.set_val(complete_key, val, units)

    def _get_complete_key(self, traj_name, phase_name):
        '''
        Compose the complete key for setting the initial guess.
        '''
        _ = traj_name
        _ = phase_name

        return self.key


class InitialGuessControl(InitialGuess):
    '''
    Define a utility for setting an initial guess for a control on a problem.

    Attributes
    ----------
    key : str
        base name for the control
    '''
    __slots__ = ()

    def _get_complete_key(self, traj_name, phase_name):
        '''
        Compose the complete key for setting the control initial guess.
        '''
        key = f'{traj_name}.{phase_name}.controls:{self.key}'

        return key


class InitialGuessParameter(InitialGuess):
    '''
    Define a utility for setting an initial guess for a parameter on a problem.

    Attributes
    ----------
    key : str
        base name for the parameter
    '''
    __slots__ = ()

    def _get_complete_key(self, traj_name, phase_name):
        '''
        Compose the complete key for setting the parameter initial guess.
        '''
        key = f'{traj_name}.{phase_name}.parameters:{self.key}'

        return key


class InitialGuessPolynomialControl(InitialGuess):
    '''
    Define a utility for setting an initial guess for a polynomial control on a problem.

    Attributes
    ----------
    key : str
        base name for the polynomial control
    '''
    __slots__ = ()

    def _get_complete_key(self, traj_name, phase_name):
        '''
        Compose the complete key for setting the polynomial control initial guess.
        '''
        key = f'{traj_name}.{phase_name}.polynomial_controls:{self.key}'

        return key


class InitialGuessState(InitialGuess):
    '''
    Define a utility for setting an initial guess for a state on a problem.

    Attributes
    ----------
    key : str
        base name for the state
    '''
    __slots__ = ()

    def _get_complete_key(self, traj_name, phase_name):
        '''
        Compose the complete key for setting the state initial guess.
        '''
        key = f'{traj_name}.{phase_name}.states:{self.key}'

        return key


class InitialGuessTime(InitialGuess):
    '''
    Define a utility for setting guesses for initial time and duration on a problem.

    Attributes
    ----------
    key : str ('times')
        the group identifier for guesses for initial time and duration
    '''
    __slots__ = ()

    def __init__(self, key='times'):
        super().__init__(key)

    def apply_initial_guess(
        self, prob: om.Problem, traj_name, phase: dm.Phase, phase_name, val, units
    ):
        '''
        Set the guesses for initial time and duration on the problem.
        '''
        _ = phase

        name = f'{traj_name}.{phase_name}.t_initial'
        t_initial, t_duration = val
        prob.set_val(name, t_initial, units)

        name = f'{traj_name}.{phase_name}.t_duration'
        prob.set_val(name, t_duration, units)



'''
Define utilities to calculate the estimated mass of any passengers, their
baggage, and other cargo. The methodology is based on the FLOPS weight
equations, modified to output mass instead of weight.
'''
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class CargoMass(om.ExplicitComponent):
    '''
    Calculate the estimated mass of any passengers, their baggage, and other
    cargo.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_output(self, Aircraft.CrewPayload.PASSENGER_MASS, 0.)
        add_aviary_output(self, Aircraft.CrewPayload.BAGGAGE_MASS, 0.)
        add_aviary_output(self, Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS, 0.)

        add_aviary_input(self, Aircraft.CrewPayload.WING_CARGO, 0.)
        add_aviary_input(self, Aircraft.CrewPayload.MISC_CARGO, 0.)

        add_aviary_output(self, Aircraft.CrewPayload.CARGO_MASS, 0.)
        add_aviary_output(self, Aircraft.CrewPayload.TOTAL_PAYLOAD_MASS, 0.)

    def setup_partials(self):

        self.declare_partials(
            Aircraft.CrewPayload.CARGO_MASS, Aircraft.CrewPayload.WING_CARGO, val=1.
        )

        self.declare_partials(
            Aircraft.CrewPayload.CARGO_MASS, Aircraft.CrewPayload.MISC_CARGO, val=1.
        )

        self.declare_partials(
            Aircraft.CrewPayload.TOTAL_PAYLOAD_MASS,
            [
                Aircraft.CrewPayload.WING_CARGO,
                Aircraft.CrewPayload.MISC_CARGO
            ],
            val=1.0
        )

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aviary_options: AviaryValues = self.options['aviary_options']
        passenger_count = aviary_options.get_val(Aircraft.CrewPayload.NUM_PASSENGERS)

        mass_per_passenger = aviary_options.get_val(
            Aircraft.CrewPayload.MASS_PER_PASSENGER, units='lbm')

        baggage_mass_per_passenger = aviary_options.get_val(
            Aircraft.CrewPayload.BAGGAGE_MASS_PER_PASSENGER, 'lbm')

        outputs[Aircraft.CrewPayload.PASSENGER_MASS] = \
            mass_per_passenger * passenger_count

        outputs[Aircraft.CrewPayload.BAGGAGE_MASS] = \
            baggage_mass_per_passenger * passenger_count

        outputs[Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS] = outputs[Aircraft.CrewPayload.PASSENGER_MASS] + \
            outputs[Aircraft.CrewPayload.BAGGAGE_MASS]

        wing_cargo = inputs[Aircraft.CrewPayload.WING_CARGO]
        misc_cargo = inputs[Aircraft.CrewPayload.MISC_CARGO]

        outputs[Aircraft.CrewPayload.CARGO_MASS] = wing_cargo + misc_cargo

        outputs[Aircraft.CrewPayload.TOTAL_PAYLOAD_MASS] = (
            outputs[Aircraft.CrewPayload.PASSENGER_MASS]
            + outputs[Aircraft.CrewPayload.BAGGAGE_MASS]
            + outputs[Aircraft.CrewPayload.CARGO_MASS])



import math

import numpy as np


def distributed_engine_count_factor(total_num_eng: int) -> float:
    """
    Returns the distributed propulsion engine count factor.

    Parameters
    ----------
    total_num_eng : int
        Total number of engines
    """

    factor = float(total_num_eng)

    if total_num_eng > 4:
        factor = 4.0 + 2.0 * math.atan((total_num_eng - 4.0) / 3.0)

    return factor


def distributed_thrust_factor(max_sls_thrust: float, total_num_eng: int) -> float:
    """
    Returns the distributed propulsion thrust factor.

    Parameters
    ----------
    max_sls_thrust : float
        Total maximum SLS thrust
    total_num_eng : iterable or int
        Total number of engines
    """

    num_engine_factor = distributed_engine_count_factor(total_num_eng)
    return max_sls_thrust / num_engine_factor


def distributed_nacelle_diam_factor(diam_nacelle: list, num_eng: list) -> float:
    """
    Returns the distributed propulsion nacelle average diameter factor.

    Parameters
    ----------
    diam_nacelle : iterable or float
        Nacelle average diameter for each engine model
    num_eng : iterable or int
        Number of engines for each engine model
    """

    # If there is more than one engine model, use the global average diameter
    try:
        total_num_eng = sum(num_eng)
        diam_avg = sum(diam_nacelle * num_eng) / total_num_eng
    except TypeError:
        total_num_eng = num_eng
        diam_avg = diam_nacelle

    diam_factor = diam_avg
    if total_num_eng > 4:
        diam_factor = 0.5 * diam_avg * total_num_eng ** 0.5

    return diam_factor


def distributed_nacelle_diam_factor_deriv(num_eng: int) -> float:
    """
    Returns the derivative of the distributed propulsion nacelle average diameter factor w.r.t.
    the global nacelle average diameter.

    Parameters
    ----------
    num_eng : iterable or int
        Number of engines for each engine model
    """

    try:
        total_num_engines = sum(num_eng)
    except TypeError:
        total_num_engines = num_eng

    deriv = 1.0
    if total_num_engines > 4:
        deriv = 0.5 * total_num_engines ** 0.5

    return deriv


def nacelle_count_factor(num_eng):
    """
    Returns the nacelle count factor, which is the number of engines plus
    0.5 if there is a centerline engine. It is assumed there is a centerline
    engine if the number of engines is odd.

    Parameters
    ----------
    num_eng : iterable or int
        Number of engines for each engine model
    """
    try:
        return np.array([n + n % 2 * 0.5 for n in num_eng])
    except TypeError:
        return np.array([num_eng + num_eng % 2 * 0.5])



import openmdao.utils.units as units

units.add_unit('distance_units', '1*m')

GRAV_METRIC_GASP = 9.81  # m/s^2
GRAV_ENGLISH_GASP = 32.2  # ft/s^2
GRAV_METRIC_FLOPS = 9.80665  # m/s^2
GRAV_ENGLISH_FLOPS = 32.17399  # ft/s^2
GRAV_ENGLISH_LBM = 1.0  # lbf/lbm
RHO_SEA_LEVEL_ENGLISH = 0.0023769  # slug/ft^3
RHO_SEA_LEVEL_METRIC = 1.225  # kg/m^3
MU_TAKEOFF = 0.02  # TODO: fill in coefficient of friction for takeoff
MU_LANDING = 0.4  # from gasp output
# sea level standard pressure in psf
PSLS_PSF = 2116.22
# sea level standard temperature in deg R
TSLS_DEGR = 518.67
RADIUS_EARTH_METRIC = 6367533.0  # m (meridional)



import unittest

from aviary.utils.test_utils.variable_test import (
    DuplicateHierarchy, assert_metadata_alphabetization, assert_no_duplicates,
    assert_structure_alphabetization, get_names_from_hierarchy)
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.variable_info.variables import Aircraft, Mission


class MetaDataTest(unittest.TestCase):
    def test_duplicate_names_FLOPS(self):

        flops_names = [var["historical_name"]['FLOPS'] for var in _MetaData.values()]
        assert_no_duplicates(flops_names)

    def test_duplicate_names_LEAPS1(self):

        leaps1_names = [var["historical_name"]['LEAPS1'] for var in _MetaData.values()]
        assert_no_duplicates(leaps1_names)

    def test_alphabetization(self):

        # TODO currently excluding Dynamic variables that do not have proper full
        #      names mirroring the hierarchy
        metadata_var_names = [key for key in _MetaData if ':' in key]

        assert_metadata_alphabetization(metadata_var_names)


class VariableStructureTest(unittest.TestCase):
    def test_duplicate_names_Aviary(self):

        aviary_names = get_names_from_hierarchy(
            Aircraft) + get_names_from_hierarchy(Mission)

        assert_no_duplicates(aviary_names)

    def test_alphabetization(self):

        assert_structure_alphabetization("variable_info/variables.py")


class TestTheTests(unittest.TestCase):
    def test_duplication_check(self):

        with self.assertRaises(ValueError) as cm:
            duplicated_names = get_names_from_hierarchy(DuplicateHierarchy)
            assert_no_duplicates(duplicated_names)
        self.assertEqual(str(
            cm.exception), "The variables ['mission:design:cruise_mach'] are duplicates in the provided list.")


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.mission.gasp_based.ode.accel_eom import AccelerationRates
from aviary.variable_info.variables import Dynamic


class AccelerationTestCase(unittest.TestCase):
    """
    These tests compare the output of the accel EOM to the output from GASP. There are some discrepancies.
    These discrepancies were considered to be small enough, given that the difference in calculation methods between the two codes is significant.
    Note that the discrepancies occur in values that had to be finite differenced.
    """

    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group", AccelerationRates(num_nodes=2), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Dynamic.Mission.MASS, np.array([174878, 174878]), units="lbm"
        )
        self.prob.model.set_input_defaults(
            Dynamic.Mission.DRAG, np.array([2635.225, 2635.225]), units="lbf"
        )  # note: this input value is not provided in the GASP data, so an estimation was made based on another similar data point
        self.prob.model.set_input_defaults(
            Dynamic.Mission.THRUST_TOTAL, np.array([32589, 32589]), units="lbf"
        )
        self.prob.model.set_input_defaults(
            Dynamic.Mission.VELOCITY, np.array([252, 252]), units="kn")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        tol = 1e-6
        self.prob.run_model()

        assert_near_equal(
            self.prob[Dynamic.Mission.VELOCITY_RATE], np.array(
                [5.51533958, 5.51533958]), tol
            # note: this was finite differenced from GASP. The fd value is: np.array([5.2353365, 5.2353365])
        )
        assert_near_equal(
            self.prob[Dynamic.Mission.DISTANCE_RATE], np.array(
                [425.32808399, 425.32808399]), tol
            # note: this was finite differenced from GASP. The fd value is: np.array([441.6439, 441.6439])
        )

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.geometry.gasp_based.wing import (WingFold, WingGroup,
                                                        WingParameters,
                                                        WingSize)
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft, Mission


class WingSizeTestCase1(
    unittest.TestCase
):  # actual GASP test case, input and output values based on large single aisle 1 v3 without bug fix
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("size", WingSize(), promotes=["*"])

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, 175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, 128, units="lbf/ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, 10.13, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 2e-4
        assert_near_equal(self.prob[Aircraft.Wing.AREA], 1370.3, tol)
        assert_near_equal(self.prob[Aircraft.Wing.SPAN], 117.8, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class WingParametersTestCase1(
    unittest.TestCase
):  # actual GASP test case, input and output values based on large single aisle 1 v3 without bug fix
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "parameters", WingParameters(aviary_options=get_option_defaults()), promotes=["*"]
        )

        self.prob.model.set_input_defaults(Aircraft.Wing.AREA, 1370.3, units="ft**2")
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, 117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, 10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, 0.33, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, 25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, 0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, 13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(self.prob[Aircraft.Wing.CENTER_CHORD], 17.49, tol)
        assert_near_equal(self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.615, tol)
        assert_near_equal(self.prob[Aircraft.Wing.ROOT_CHORD], 16.41, tol)
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.1397, tol
        )  # this is slightly different from the GASP output value, likely due to rounding error
        assert_near_equal(self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 1114, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class WingParametersTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "parameters", WingParameters(aviary_options=options), promotes=["*"]
        )

        self.prob.model.set_input_defaults(Aircraft.Wing.AREA, 1370.3, units="ft**2")
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, 117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, 10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, 0.33, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, 25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, 0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, 13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(self.prob[Aircraft.Wing.CENTER_CHORD], 17.49, tol)
        assert_near_equal(self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.615, tol)
        assert_near_equal(self.prob[Aircraft.Wing.ROOT_CHORD], 16.41, tol)
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.1397, tol
        )  # this is slightly different from the GASP output value, likely due to rounding error

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class WingFoldTestCase1(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, val=False, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingFold(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, 0.33, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, 0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )
        self.prob.model.set_input_defaults(
            "strut_y", val=25, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4

        assert_near_equal(
            self.prob["nonfolded_taper_ratio"], 0.71561969, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDING_AREA], 620.04352246, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["nonfolded_wing_area"], 750.25647754, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["tc_ratio_mean_folded"], 0.14363328, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["nonfolded_AR"], 3.33219382, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 712.3428037422319, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class WingFoldTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingFold(
                aviary_options=options
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, 0.33, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, 0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN, val=25, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4

        assert_near_equal(
            self.prob["nonfolded_taper_ratio"], 0.85780985, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDING_AREA], 964.0812219, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["nonfolded_wing_area"], 406.2187781, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["tc_ratio_mean_folded"], 0.14681664, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["nonfolded_AR"], 1.53857978, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 406.64971668264957, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-12, rtol=1e-12)


class WingGroupTestCase1(
    unittest.TestCase
):  # actual GASP test case, input and output values based on large single aisle 1 v3 without bug fix
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group", WingGroup(aviary_options=get_option_defaults()), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, 175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, 128, units="lbf/ft**2"
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, 10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, 0.33, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, 25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, 0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, 13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4

        assert_near_equal(self.prob[Aircraft.Wing.AREA], 1370.3, tol)
        assert_near_equal(self.prob[Aircraft.Wing.SPAN], 117.8, tol)

        assert_near_equal(self.prob[Aircraft.Wing.CENTER_CHORD], 17.49, tol)
        assert_near_equal(self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.615, tol)
        assert_near_equal(self.prob[Aircraft.Wing.ROOT_CHORD], 16.41, tol)
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.1397, tol
        )  # this is slightly different from the GASP output value, likely due to rounding error
        assert_near_equal(self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 1114, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-12, rtol=1e-12)


class WingGroupTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, val=False, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, 175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, 128, units="lbf/ft**2"
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, 10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, 0.33, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, 25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, 0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, 13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.Strut.AREA_RATIO, val=0.02189, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION, val=1.0, units="ft"
        )  # not actual GASP value

        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4

        assert_near_equal(self.prob[Aircraft.Wing.AREA], 1370.3, tol)
        assert_near_equal(self.prob[Aircraft.Wing.SPAN], 117.8, tol)

        assert_near_equal(self.prob[Aircraft.Wing.CENTER_CHORD], 17.49, tol)
        assert_near_equal(self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.615, tol)
        assert_near_equal(self.prob[Aircraft.Wing.ROOT_CHORD], 16.41, tol)
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.1397, tol
        )  # this is slightly different from the GASP output value, likely due to rounding error

        assert_near_equal(
            self.prob["fold.nonfolded_taper_ratio"], 0.9943133, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDING_AREA], 1352.8724859, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fold.nonfolded_wing_area"], 17.4400141, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fold.tc_ratio_mean_folded"], 0.14987269, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fold.nonfolded_AR"], 0.0573394, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 18.26837098, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Strut.LENGTH], 14.42957033, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.CHORD], 1.03953199, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=7e-12, rtol=1e-12)


class WingGroupTestCase3(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, 175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, 128, units="lbf/ft**2"
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, 10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, 0.33, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, 25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, 0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, 13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN, val=25, units="ft"
        )  # not actual GASP value

        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4

        assert_near_equal(self.prob[Aircraft.Wing.AREA], 1370.3, tol)
        assert_near_equal(self.prob[Aircraft.Wing.SPAN], 117.8, tol)

        assert_near_equal(self.prob[Aircraft.Wing.CENTER_CHORD], 17.49, tol)
        assert_near_equal(self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.615, tol)
        assert_near_equal(self.prob[Aircraft.Wing.ROOT_CHORD], 16.41, tol)
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.1397, tol
        )  # this is slightly different from the GASP output value, likely due to rounding error

        assert_near_equal(
            self.prob["fold.nonfolded_taper_ratio"], 0.85780985, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDING_AREA], 964.14982163, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fold.nonfolded_wing_area"], 406.16267837, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fold.tc_ratio_mean_folded"], 0.14681715, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fold.nonfolded_AR"], 1.5387923, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 406.64971668264957, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-12, rtol=1e-12)


class WingGroupTestCase4(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN, val=1, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION, val=0, units="ft"
        )

        self.prob.setup(check=False, force_alloc_complex=True)


class WingGroupTestCase5(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, 175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, 128, units="lbf/ft**2"
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, 10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, 0.33, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, 25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, 0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, 13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.Strut.AREA_RATIO, val=.2, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=150, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION, val=1.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.AREA_RATIO, val=0.021893, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4

        assert_near_equal(self.prob[Aircraft.Wing.AREA], 1370.3, tol)
        assert_near_equal(self.prob[Aircraft.Wing.SPAN], 117.8, tol)

        assert_near_equal(self.prob[Aircraft.Wing.CENTER_CHORD], 17.49, tol)
        assert_near_equal(self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.615, tol)
        assert_near_equal(self.prob[Aircraft.Wing.ROOT_CHORD], 16.41, tol)
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.1397, tol
        )  # this is slightly different from the GASP output value, likely due to rounding error

        assert_near_equal(
            self.prob[Aircraft.Strut.LENGTH], 14.42957033, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.CHORD], 1.03953199, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import copy
import unittest
import warnings

import dymos as dm
import openmdao.api as om
from openmdao.core.driver import Driver
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs

from aviary.interface.methods_for_level2 import AviaryProblem

from aviary.utils.functions import set_aviary_initial_values
from aviary.utils.preprocessors import preprocess_crewpayload
from aviary.models.N3CC.N3CC_data import \
    balanced_liftoff_user_options as _takeoff_liftoff_user_options
from aviary.models.N3CC.N3CC_data import \
    balanced_trajectory_builder as _takeoff_trajectory_builder
from aviary.models.N3CC.N3CC_data import \
    inputs as _inputs
from aviary.variable_info.variables import Dynamic
from aviary.variable_info.variables_in import VariablesIn
from aviary.interface.default_phase_info.height_energy import default_premission_subsystems
from aviary.subsystems.premission import CorePreMission


@use_tempdirs
class TestFLOPSBalancedFieldLength(unittest.TestCase):
    @require_pyoptsparse(optimizer='IPOPT')
    def bench_test_IPOPT(self):
        # raise unittest.SkipTest("IPOPT currently not working with this benchmark.")

        driver = om.pyOptSparseDriver()

        optimizer = 'IPOPT'
        driver.options['optimizer'] = optimizer

        driver.opt_settings['max_iter'] = 100
        driver.opt_settings['tol'] = 1e-3
        driver.opt_settings['print_level'] = 4

        self._do_run(driver, optimizer)

    @require_pyoptsparse(optimizer='SNOPT')
    def bench_test_SNOPT(self):
        driver = om.pyOptSparseDriver()

        optimizer = 'SNOPT'
        driver.options['optimizer'] = optimizer

        driver.opt_settings['Major iterations limit'] = 50
        driver.opt_settings['Major optimality tolerance'] = 1e-4
        driver.opt_settings['Major feasibility tolerance'] = 1e-6
        driver.opt_settings['iSumm'] = 6

        self._do_run(driver, optimizer)

    def _do_run(self, driver: Driver, optimizer, *args):
        aviary_options = _inputs.deepcopy()
        preprocess_crewpayload(aviary_options)

        takeoff_trajectory_builder = copy.deepcopy(_takeoff_trajectory_builder)
        takeoff_liftoff_user_options = _takeoff_liftoff_user_options.deepcopy()

        takeoff = om.Problem()
        takeoff.driver = driver

        driver.declare_coloring()

        driver.add_recorder(om.SqliteRecorder(
            f'FLOPS_detailed_takeoff_traj_{optimizer}.sql'))

        driver.recording_options['record_derivatives'] = False

        # Upstream static analysis for aero
        takeoff.model.add_subsystem(
            'core_subsystems',
            CorePreMission(
                aviary_options=aviary_options,
                subsystems=default_premission_subsystems,
            ),
            promotes_inputs=['aircraft:*', 'mission:*'],
            promotes_outputs=['aircraft:*', 'mission:*'])

        # Instantiate the trajectory and add the phases
        traj = dm.Trajectory()
        takeoff.model.add_subsystem('traj', traj)

        takeoff_trajectory_builder.build_trajectory(
            aviary_options=aviary_options, model=takeoff.model, traj=traj)

        distance_max, units = takeoff_liftoff_user_options.get_item('distance_max')
        liftoff = takeoff_trajectory_builder.get_phase('balanced_liftoff')

        liftoff.add_objective(
            Dynamic.Mission.DISTANCE, loc='final', ref=distance_max, units=units)

        takeoff.model.add_subsystem(
            'input_sink',
            VariablesIn(aviary_options=aviary_options),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        # suppress warnings:
        # "input variable '...' promoted using '*' was already promoted using 'aircraft:*'
        with warnings.catch_warnings():
            # Set initial default values for all aircraft variables.
            set_aviary_initial_values(takeoff.model, aviary_options)

            warnings.simplefilter("ignore", om.PromotionWarning)
            takeoff.setup(check=True)

        # Turn off solver printing so that the SNOPT output is readable.
        takeoff.set_solver_print(level=0)

        takeoff_trajectory_builder.apply_initial_guesses(takeoff, 'traj')

        # run the problem
        dm.run_problem(takeoff, run_driver=True, simulate=True, make_plots=False)

        # takeoff.model.traj.phases.brake_release_to_decision_speed.list_inputs(print_arrays=True)
        # takeoff.model.list_outputs(print_arrays=True)

        # Field Length
        # N3CC FLOPS output Line 2282
        desired = 7032.65
        actual = takeoff.model.get_val(
            'traj.balanced_liftoff.states:distance', units='ft')[-1]
        assert_near_equal(actual, desired, 2e-2)

        # Decision Time
        # N3CC FLOPS output Line 2287
        desired = 29.52
        actual = takeoff.model.get_val(
            'traj.balanced_brake_release.t', units='s')[-1]
        assert_near_equal(actual, desired, 2e-2)

        # Liftoff Time
        # N3CC FLOPS output Line 2289
        desired = 36.63
        actual = takeoff.model.get_val(
            'traj.balanced_rotate.t', units='s')[-1]
        assert_near_equal(actual, desired, .05)

        # Rotation Speed
        # N3CC FLOPS output Line 2289
        desired = 156.55
        actual = takeoff.model.get_val(
            'traj.balanced_rotate.states:velocity', units='kn')[-1]
        assert_near_equal(actual, desired, 2e-2)


if __name__ == '__main__':
    use_SNOPT = True

    z = TestFLOPSBalancedFieldLength()

    if use_SNOPT:
        z.bench_test_SNOPT()

    else:
        z.bench_test_IPOPT()



import numpy as np
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.mission.gasp_based.ode.accel_eom import AccelerationRates
from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.subsystems.mass.mass_to_weight import MassToWeight
from aviary.variable_info.enums import AnalysisScheme, SpeedType
from aviary.variable_info.variables import Aircraft, Dynamic, Mission
from aviary.mission.ode.specific_energy_rate import SpecificEnergyRate
from aviary.mission.ode.altitude_rate import AltitudeRate


class AccelODE(BaseODE):
    """ODE for level acceleration.

    In level acceleration, there are only nonzero net forces in the direction of motion.
    There is a balance component to solve for the angle of attack necessary to make lift
    equal to weight. Acceleration results from engine thrust in excess of drag.
    """

    def setup(self):
        nn = self.options["num_nodes"]
        analysis_scheme = self.options["analysis_scheme"]
        aviary_options = self.options['aviary_options']
        core_subsystems = self.options['core_subsystems']

        # TODO: paramport
        self.add_subsystem("params", ParamPort(), promotes=["*"])

        self.add_subsystem(
            "USatm",
            USatm1976Comp(
                num_nodes=nn),
            promotes_inputs=[
                ("h",
                 Dynamic.Mission.ALTITUDE)],
            promotes_outputs=[
                "rho",
                ("sos",
                 Dynamic.Mission.SPEED_OF_SOUND),
                ("temp",
                 Dynamic.Mission.TEMPERATURE),
                ("pres",
                 Dynamic.Mission.STATIC_PRESSURE),
                "viscosity"],
        )

        self.add_flight_conditions(nn)

        self.add_subsystem(
            "calc_weight",
            MassToWeight(num_nodes=nn),
            promotes_inputs=[("mass", Dynamic.Mission.MASS)],
            promotes_outputs=["weight"]
        )

        kwargs = {'num_nodes': nn, 'aviary_inputs': aviary_options,
                  'method': 'cruise', 'output_alpha': True}
        for subsystem in core_subsystems:
            system = subsystem.build_mission(**kwargs)
            if system is not None:
                self.add_subsystem(subsystem.name,
                                   system,
                                   promotes_inputs=subsystem.mission_inputs(**kwargs),
                                   promotes_outputs=subsystem.mission_outputs(**kwargs))

        sgm_inputs = [
            't_curr', Dynamic.Mission.DISTANCE] if analysis_scheme is AnalysisScheme.SHOOTING else []
        sgm_outputs = [
            Dynamic.Mission.ALTITUDE_RATE] if analysis_scheme is AnalysisScheme.SHOOTING else []

        self.add_subsystem(
            "eom",
            AccelerationRates(
                num_nodes=nn,
                analysis_scheme=analysis_scheme),
            promotes_inputs=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.VELOCITY,
                Dynamic.Mission.DRAG,
                Dynamic.Mission.THRUST_TOTAL, ]
            + sgm_inputs,
            promotes_outputs=[
                Dynamic.Mission.VELOCITY_RATE,
                Dynamic.Mission.DISTANCE_RATE, ]
            + sgm_outputs,
        )

        self.add_excess_rate_comps(nn)

        ParamPort.set_default_vals(self)
        self.set_input_defaults(Dynamic.Mission.MASS, val=14e4 *
                                np.ones(nn), units="lbm")
        self.set_input_defaults(Dynamic.Mission.ALTITUDE,
                                val=500 * np.ones(nn), units="ft")
        self.set_input_defaults(Dynamic.Mission.VELOCITY, val=200*np.ones(nn),
                                units="m/s")  # val here is nominal



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.variable_info.enums import AnalysisScheme
from aviary.variable_info.variables import Dynamic


class ClimbRates(om.ExplicitComponent):

    """
    Compute the altitude rate, distance rate, required lift, and flight path angle for
    an aircraft in a climb phase of flight.

    Equations come from climb subroutine of GASP code.
    """

    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare("analysis_scheme", types=AnalysisScheme, default=AnalysisScheme.COLLOCATION,
                             desc="The analysis method that will be used to close the trajectory; for example collocation or time integration")

    def setup(self):
        analysis_scheme = self.options["analysis_scheme"]
        nn = self.options["num_nodes"]
        arange = np.arange(nn)

        self.add_input(
            Dynamic.Mission.VELOCITY,
            val=np.zeros(nn),
            units="ft/s",
            desc="true air speed",
        )

        self.add_input(Dynamic.Mission.THRUST_TOTAL, val=np.zeros(nn),
                       units="lbf", desc="net thrust")
        self.add_input(
            Dynamic.Mission.DRAG,
            val=np.zeros(nn),
            units="lbf",
            desc="net drag on aircraft")
        self.add_input(
            Dynamic.Mission.MASS,
            val=np.zeros(nn),
            units="lbm",
            desc="mass of aircraft",
        )

        self.add_output(
            Dynamic.Mission.ALTITUDE_RATE,
            val=np.zeros(nn),
            units="ft/s",
            desc="rate of change of altitude",
        )
        self.add_output(
            Dynamic.Mission.DISTANCE_RATE,
            val=np.zeros(nn),
            units="ft/s",
            desc="rate of change of horizontal distance covered",
        )
        self.add_output(
            "required_lift",
            val=np.zeros(nn),
            units="lbf",
            desc="lift required in order to maintain calculated flight path angle",
        )
        self.add_output(
            Dynamic.Mission.FLIGHT_PATH_ANGLE,
            val=np.ones(nn),
            units="rad",
            desc="flight path angle",
        )

        self.declare_partials(Dynamic.Mission.ALTITUDE_RATE,
                              [Dynamic.Mission.VELOCITY,
                               Dynamic.Mission.THRUST_TOTAL,
                               Dynamic.Mission.DRAG,
                               Dynamic.Mission.MASS],
                              rows=arange,
                              cols=arange)
        self.declare_partials(
            Dynamic.Mission.DISTANCE_RATE,
            [Dynamic.Mission.VELOCITY, Dynamic.Mission.THRUST_TOTAL,
                Dynamic.Mission.DRAG, Dynamic.Mission.MASS],
            rows=arange,
            cols=arange,
        )
        self.declare_partials("required_lift",
                              [Dynamic.Mission.MASS,
                               Dynamic.Mission.THRUST_TOTAL,
                               Dynamic.Mission.DRAG],
                              rows=arange,
                              cols=arange)
        self.declare_partials(Dynamic.Mission.FLIGHT_PATH_ANGLE,
                              [Dynamic.Mission.THRUST_TOTAL,
                               Dynamic.Mission.DRAG,
                               Dynamic.Mission.MASS],
                              rows=arange,
                              cols=arange)

        if analysis_scheme is AnalysisScheme.SHOOTING:
            self.add_input("t_curr", val=np.ones(nn), desc="time", units="s")
            self.add_input(
                Dynamic.Mission.DISTANCE, val=np.ones(nn), desc="distance traveled", units="ft"
            )

    def compute(self, inputs, outputs):

        TAS = inputs[Dynamic.Mission.VELOCITY]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        drag = inputs[Dynamic.Mission.DRAG]
        weight = inputs[Dynamic.Mission.MASS] * GRAV_ENGLISH_LBM

        gamma = np.arcsin((thrust - drag) / weight)

        outputs[Dynamic.Mission.ALTITUDE_RATE] = TAS * np.sin(gamma)
        outputs[Dynamic.Mission.DISTANCE_RATE] = TAS * np.cos(gamma)
        outputs["required_lift"] = weight * np.cos(gamma)
        outputs[Dynamic.Mission.FLIGHT_PATH_ANGLE] = gamma

    def compute_partials(self, inputs, J):

        TAS = inputs[Dynamic.Mission.VELOCITY]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        drag = inputs[Dynamic.Mission.DRAG]
        weight = inputs[Dynamic.Mission.MASS] * GRAV_ENGLISH_LBM

        gamma = np.arcsin((thrust - drag) / weight)

        dGamma_dThrust = (1 - ((thrust - drag) / weight) ** 2) ** (-0.5) / weight
        dGamma_dDrag = (1 - ((thrust - drag) / weight) ** 2) ** (-0.5) / (-weight)
        dGamma_dWeight = (
            (1 - ((thrust - drag) / weight) ** 2) ** (-0.5)
            * (drag - thrust)
            / weight**2
        )

        J[Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.VELOCITY] = np.sin(gamma)
        J[Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.THRUST_TOTAL] = TAS * \
            np.cos(gamma) * dGamma_dThrust
        J[Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.DRAG] = TAS * \
            np.cos(gamma) * dGamma_dDrag
        J[Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.MASS] = \
            TAS * np.cos(gamma) * dGamma_dWeight * GRAV_ENGLISH_LBM

        J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.VELOCITY] = np.cos(gamma)
        J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.THRUST_TOTAL] = - \
            TAS * np.sin(gamma) * dGamma_dThrust
        J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.DRAG] = - \
            TAS * np.sin(gamma) * dGamma_dDrag
        J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.MASS] = \
            -TAS * np.sin(gamma) * dGamma_dWeight * GRAV_ENGLISH_LBM

        J["required_lift", Dynamic.Mission.MASS] = (
            np.cos(gamma) - weight * np.sin(gamma) * dGamma_dWeight
        ) * GRAV_ENGLISH_LBM
        J["required_lift", Dynamic.Mission.THRUST_TOTAL] = - \
            weight * np.sin(gamma) * dGamma_dThrust
        J["required_lift", Dynamic.Mission.DRAG] = -weight * np.sin(gamma) * dGamma_dDrag

        J[Dynamic.Mission.FLIGHT_PATH_ANGLE, Dynamic.Mission.THRUST_TOTAL] = dGamma_dThrust
        J[Dynamic.Mission.FLIGHT_PATH_ANGLE, Dynamic.Mission.DRAG] = dGamma_dDrag
        J[Dynamic.Mission.FLIGHT_PATH_ANGLE,
            Dynamic.Mission.MASS] = dGamma_dWeight * GRAV_ENGLISH_LBM



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.landing_gear import (
    AltLandingGearMass, LandingGearMass, MainGearLength, NoseGearLength)
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (Version,
                                                      flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Mission


class LandingGearMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            "landing_gear",
            LandingGearMass(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            self.prob,
            case_name,
            input_keys=[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                        Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER,
                        Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH,
                        Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER,
                        Aircraft.Design.TOUCHDOWN_MASS],
            output_keys=[Aircraft.LandingGear.MAIN_GEAR_MASS,
                         Aircraft.LandingGear.NOSE_GEAR_MASS],
            version=Version.TRANSPORT,
            atol=1e-11)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class AltLandingGearMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            "landing_gear_alt",
            AltLandingGearMass(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            self.prob,
            case_name,
            input_keys=[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                        Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER,
                        Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH,
                        Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER,
                        Mission.Design.GROSS_MASS],
            output_keys=[Aircraft.LandingGear.MAIN_GEAR_MASS,
                         Aircraft.LandingGear.NOSE_GEAR_MASS],
            version=Version.ALTERNATE,
            atol=1e-11)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class LandingGearLengthTest(unittest.TestCase):
    """
    This component is unrepresented in our test data.
    """

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(only='N3CC'),
                          name_func=print_case)
    def test_derivs(self, case_name):
        prob = self.prob
        model = prob.model
        flops_inputs = get_flops_inputs(case_name)

        model.add_subsystem(
            'main', MainGearLength(aviary_options=flops_inputs), promotes=['*'])
        model.add_subsystem(
            'nose', NoseGearLength(aviary_options=flops_inputs), promotes=['*'])

        prob.setup(force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Fuselage.LENGTH,
                        Aircraft.Fuselage.MAX_WIDTH,
                        Aircraft.Nacelle.AVG_DIAMETER,
                        Aircraft.Engine.WING_LOCATIONS,
                        Aircraft.Wing.DIHEDRAL,
                        Aircraft.Wing.SPAN],
            output_keys=[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                         Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH],
            version=Version.ALTERNATE,
            atol=1e-11)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials

from aviary.subsystems.aerodynamics.flops_based.compressibility_drag import \
    CompressibilityDrag
from aviary.variable_info.variables import Aircraft, Mission


class CompressibilityDragTest(unittest.TestCase):

    def test_derivs(self):

        # Nudge the mach and diam/wingspan values off of the table points to prevent problem with
        # linear interp.
        delta = 1e-5

        mach = [.2, .25, .3, .35, .4, .45, .5, .55, .6, .7, .75, .775, .8, .825, .85, .875, 0.9,
                0.95, 1.05, 1.1]
        mach = np.array(mach) + delta
        nn = len(mach)

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('LID', CompressibilityDrag(num_nodes=nn),
                            promotes_inputs=[Aircraft.Design.BASE_AREA,
                                             Aircraft.Wing.AREA,
                                             Aircraft.Wing.ASPECT_RATIO,
                                             Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN,
                                             Aircraft.Wing.SWEEP,
                                             Aircraft.Wing.TAPER_RATIO,
                                             Aircraft.Wing.THICKNESS_TO_CHORD,
                                             Aircraft.Fuselage.CROSS_SECTION,
                                             Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
                                             Aircraft.Fuselage.LENGTH_TO_DIAMETER,
                                             Mission.Design.MACH])

        prob.setup(force_alloc_complex=True)

        prob.set_val('LID.mach', mach)
        prob.set_val(Mission.Design.MACH, .8-delta)
        prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD, 0.13)
        prob.set_val(Aircraft.Fuselage.CROSS_SECTION, 128.2)
        prob.set_val(Aircraft.Design.BASE_AREA, 0.01)
        prob.set_val(Aircraft.Wing.AREA, 1370.0)
        prob.set_val(Aircraft.Wing.TAPER_RATIO, 0.432)
        prob.set_val(Aircraft.Wing.ASPECT_RATIO, 11.5)
        prob.set_val(Aircraft.Wing.SWEEP, 25.07)
        prob.set_val(Aircraft.Fuselage.DIAMETER_TO_WING_SPAN, 0.15 + delta)
        prob.set_val(Aircraft.Fuselage.LENGTH_TO_DIAMETER, 10.12345)
        prob.set_val(Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, 0.)

        prob.run_model()
        derivs = prob.check_partials(out_stream=None, method="cs")

        # TODO: need to test outputs too
        assert_check_partials(derivs, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



"""
Define subsystem builder for Aviary core mass.

Classes
-------
MassBuilderBase : the interface for a mass subsystem builder.

CoreMassBuilder : the interface for Aviary's core mass subsystem builder
"""
from aviary.interface.utils.markdown_utils import write_markdown_variable_table
from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.subsystems.mass.flops_based.mass_premission import MassPremission as MassPremissionFLOPS
from aviary.subsystems.mass.gasp_based.mass_premission import MassPremission as MassPremissionGASP
from aviary.variable_info.variables import Aircraft, Mission
from aviary.variable_info.enums import LegacyCode


GASP = LegacyCode.GASP
FLOPS = LegacyCode.FLOPS

_default_name = 'mass'


class MassBuilderBase(SubsystemBuilderBase):
    def __init__(self, name=None, meta_data=None):
        if name is None:
            name = _default_name

        super().__init__(name=name, meta_data=meta_data)

    def mission_inputs(self, **kwargs):
        return ['*']

    def mission_outputs(self, **kwargs):
        return ['*']


class CoreMassBuilder(MassBuilderBase):
    def __init__(self, name=None, meta_data=None, code_origin=None):
        if name is None:
            name = 'core_mass'

        if code_origin not in (FLOPS, GASP):
            raise ValueError('Code origin is not one of the following: (FLOPS, GASP)')

        self.code_origin = code_origin

        super().__init__(name=name, meta_data=meta_data)

    def build_pre_mission(self, aviary_inputs):
        code_origin = self.code_origin

        if code_origin is GASP:
            mass_premission = MassPremissionGASP(aviary_options=aviary_inputs,)

        elif code_origin is FLOPS:
            mass_premission = MassPremissionFLOPS(aviary_options=aviary_inputs)

        return mass_premission

    def build_mission(self, num_nodes, aviary_inputs, **kwargs):
        super().build_mission(num_nodes, aviary_inputs)

    def report(self, prob, reports_folder, **kwargs):
        """
        Generate the report for Aviary core mass

        Parameters
        ----------
        prob : AviaryProblem
            The AviaryProblem that will be used to generate the report
        reports_folder : Path
            Location of the subsystems_report folder this report will be placed in
        """
        filename = self.name + '.md'
        filepath = reports_folder / filename

        # TODO break out table into constituent categories (SAWE standard??)
        outputs = [Aircraft.Wing.MASS,
                   Aircraft.HorizontalTail.MASS,
                   Aircraft.VerticalTail.MASS,
                   Aircraft.Fins.MASS,
                   Aircraft.Canard.MASS,
                   Aircraft.Fuselage.MASS,
                   Aircraft.LandingGear.TOTAL_MASS,
                   Aircraft.Nacelle.MASS,
                   Aircraft.Design.STRUCTURE_MASS,
                   Aircraft.Propulsion.TOTAL_ENGINE_MASS,
                   Aircraft.Propulsion.TOTAL_THRUST_REVERSERS_MASS,
                   Aircraft.Propulsion.TOTAL_MISC_MASS,
                   Aircraft.Fuel.FUEL_SYSTEM_MASS,
                   Aircraft.Propulsion.MASS,
                   Aircraft.Controls.TOTAL_MASS,
                   Aircraft.APU.MASS,
                   Aircraft.Instruments.MASS,
                   Aircraft.Hydraulics.MASS,
                   Aircraft.Electrical.MASS,
                   Aircraft.Avionics.MASS,
                   Aircraft.Furnishings.MASS,
                   Aircraft.AirConditioning.MASS,
                   Aircraft.AntiIcing.MASS,
                   Aircraft.Design.SYSTEMS_EQUIP_MASS,
                   Aircraft.Design.EXTERNAL_SUBSYSTEMS_MASS,
                   Aircraft.Design.EMPTY_MASS,
                   Aircraft.CrewPayload.FLIGHT_CREW_MASS,
                   Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS,
                   Aircraft.Fuel.UNUSABLE_FUEL_MASS,
                   Aircraft.Propulsion.TOTAL_ENGINE_OIL_MASS,
                   Aircraft.CrewPayload.PASSENGER_SERVICE_MASS,
                   Aircraft.CrewPayload.CARGO_CONTAINER_MASS,
                   Aircraft.Design.OPERATING_MASS,
                   Aircraft.CrewPayload.PASSENGER_MASS,
                   Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS,
                   Aircraft.CrewPayload.CARGO_MASS,
                   Aircraft.Design.ZERO_FUEL_MASS,
                   Mission.Design.FUEL_MASS,
                   Mission.Summary.TOTAL_FUEL_MASS,
                   Mission.Summary.GROSS_MASS
                   ]

        with open(filepath, mode='w') as f:
            method = self.code_origin + ' ESTIMATING RELATIONS'
            f.write(f'# MASS ESTIMATION: {method}')
            write_markdown_variable_table(f, prob, outputs, self.meta_data)



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


def sigX(x):
    sig = 1 / (1 + np.exp(-x))

    return sig


def dSigXdX(x):
    derivative = -1 / (1 + np.exp(-x)) ** 2 * (-1 * np.exp(-x))

    return derivative


class FuselageParameters(om.ExplicitComponent):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Fuselage.DELTA_DIAMETER, val=4.5)
        add_aviary_input(self, Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, val=9.5)

        add_aviary_output(self, Aircraft.Fuselage.AVG_DIAMETER, val=0, units='inch')
        self.add_output("cabin_height", val=0, units="ft", desc="HC: height of cabin")
        self.add_output("cabin_len", val=0, units="ft", desc="LC: length of cabin")
        self.add_output("nose_height", val=0, units="ft", desc="HN: height of nose")

        self.declare_partials(
            "cabin_height",
            [
                Aircraft.Fuselage.DELTA_DIAMETER,
            ],
        )
        self.declare_partials(
            "nose_height",
            [
                Aircraft.Fuselage.DELTA_DIAMETER,
            ],
        )

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        seats_abreast = aviary_options.get_val(Aircraft.Fuselage.NUM_SEATS_ABREAST)
        seat_width = aviary_options.get_val(Aircraft.Fuselage.SEAT_WIDTH, units='inch')
        num_aisle = aviary_options.get_val(Aircraft.Fuselage.NUM_AISLES)
        aisle_width = aviary_options.get_val(Aircraft.Fuselage.AISLE_WIDTH, units='inch')
        PAX = self.options['aviary_options'].get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')
        seat_pitch = aviary_options.get_val(Aircraft.Fuselage.SEAT_PITCH, units='inch')
        delta_diameter = inputs[Aircraft.Fuselage.DELTA_DIAMETER]

        cabin_width = seats_abreast * seat_width + num_aisle * aisle_width + 12

        if PAX < 1:
            print("Warning: you have not specified at least one passenger")

        # single seat across
        cabin_len_a = PAX * seat_pitch / 12
        nose_height_a = cabin_width / 12
        cabin_height_a = nose_height_a + delta_diameter

        # multiple seats across
        cabin_len_b = (PAX - 1) * seat_pitch / (seats_abreast * 12)
        cabin_height_b = cabin_width / 12
        nose_height_b = cabin_height_b - delta_diameter

        outputs[Aircraft.Fuselage.AVG_DIAMETER] = cabin_width
        # There are separate equations for aircraft with a single seat per row vs. multiple seats per row.
        # Here and in compute_partials, these equations are smoothed using a sigmoid fnuction centered at
        # 1.5 seats, the sigmoid function is steep enough that there should be no noticable difference
        # between the smoothed function and the stepwise function at 1 and 2 seats.
        outputs["cabin_height"] = cabin_height_a * sigX(100*(1.5-seats_abreast)) + \
            cabin_height_b*sigX(100*(seats_abreast-1.5))
        outputs["cabin_len"] = cabin_len_a * sigX(100*(1.5-seats_abreast)) + \
            cabin_len_b*sigX(100*(seats_abreast-1.5))
        outputs["nose_height"] = nose_height_a * sigX(100*(1.5-seats_abreast)) + \
            nose_height_b*sigX(100*(seats_abreast-1.5))

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        seats_abreast = aviary_options.get_val(Aircraft.Fuselage.NUM_SEATS_ABREAST)

        J["nose_height", Aircraft.Fuselage.DELTA_DIAMETER] = sigX(
            100*(seats_abreast-1.5))*(-1)
        J["cabin_height", Aircraft.Fuselage.DELTA_DIAMETER] = sigX(
            100*(1.5-seats_abreast))*1


class FuselageSize(om.ExplicitComponent):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Fuselage.NOSE_FINENESS, val=1)
        self.add_input("nose_height", val=0, units="ft", desc="HN: height of nose")
        add_aviary_input(self, Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, val=9.5)
        self.add_input("cabin_len", val=0, units="ft", desc="LC: length of cabin")
        add_aviary_input(self, Aircraft.Fuselage.TAIL_FINENESS, val=3)
        self.add_input("cabin_height", val=0, units="ft", desc="HC: height of cabin")
        if self.options['aviary_options'].get_val(Aircraft.Fuselage.PROVIDE_SURFACE_AREA, units='unitless'):
            add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA_FACTOR,
                             val=4000, units="unitless")
        else:
            add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA_FACTOR,
                             val=1, units="unitless")

        add_aviary_output(self, Aircraft.Fuselage.LENGTH, val=0)
        add_aviary_output(self, Aircraft.Fuselage.WETTED_AREA, val=0)
        add_aviary_output(self, Aircraft.TailBoom.LENGTH, val=0)

        self.declare_partials(
            Aircraft.Fuselage.LENGTH,
            [
                Aircraft.Fuselage.NOSE_FINENESS,
                "nose_height",
                Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH,
                "cabin_len",
                Aircraft.Fuselage.TAIL_FINENESS,
                "cabin_height",
            ],
        )
        self.declare_partials(
            Aircraft.Fuselage.WETTED_AREA,
            [
                "cabin_height",
                Aircraft.Fuselage.NOSE_FINENESS,
                "nose_height",
                Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH,
                "cabin_len",
                Aircraft.Fuselage.TAIL_FINENESS,
            ],
        )

        if self.options['aviary_options'].get_val(Aircraft.Fuselage.PROVIDE_SURFACE_AREA, units='unitless'):
            self.declare_partials(Aircraft.Fuselage.WETTED_AREA,
                                  Aircraft.Fuselage.WETTED_AREA_FACTOR)

        else:
            self.declare_partials(
                Aircraft.Fuselage.WETTED_AREA,
                [
                    Aircraft.Fuselage.WETTED_AREA_FACTOR,
                    "cabin_height",
                    Aircraft.Fuselage.NOSE_FINENESS,
                    "nose_height",
                    Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH,
                    "cabin_len",
                    Aircraft.Fuselage.TAIL_FINENESS,
                ],
            )

        self.declare_partials(
            Aircraft.TailBoom.LENGTH,
            [
                Aircraft.Fuselage.NOSE_FINENESS,
                "nose_height",
                Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH,
                "cabin_len",
                Aircraft.Fuselage.TAIL_FINENESS,
                "cabin_height",
            ],
        )

    def compute(self, inputs, outputs):
        LoverD_nose = inputs[Aircraft.Fuselage.NOSE_FINENESS]
        LoverD_tail = inputs[Aircraft.Fuselage.TAIL_FINENESS]
        cockpit_len = inputs[Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH]
        fus_SA_factor = inputs[Aircraft.Fuselage.WETTED_AREA_FACTOR]
        nose_height = inputs["nose_height"]
        cabin_len = inputs["cabin_len"]
        cabin_height = inputs["cabin_height"]

        fus_len = (
            LoverD_nose * nose_height
            + cockpit_len
            + cabin_len
            + LoverD_tail * cabin_height
        )

        fus_SA = cabin_height * (
            2.5 * (LoverD_nose * nose_height + cockpit_len)
            + 3.14 * cabin_len
            + 2.1 * LoverD_tail * cabin_height
        )

        if self.options["aviary_options"].get_val(Aircraft.Fuselage.PROVIDE_SURFACE_AREA, units='unitless'):
            fus_SA = fus_SA_factor

        else:
            fus_SA = fus_SA * fus_SA_factor

        cabin_len_tailboom = fus_len

        outputs[Aircraft.Fuselage.LENGTH] = fus_len
        outputs[Aircraft.Fuselage.WETTED_AREA] = fus_SA
        outputs[Aircraft.TailBoom.LENGTH] = cabin_len_tailboom

    def compute_partials(self, inputs, J):
        LoverD_nose = inputs[Aircraft.Fuselage.NOSE_FINENESS]
        LoverD_tail = inputs[Aircraft.Fuselage.TAIL_FINENESS]
        nose_height = inputs["nose_height"]
        cabin_height = inputs["cabin_height"]
        fus_SA_factor = inputs[Aircraft.Fuselage.WETTED_AREA_FACTOR]
        cockpit_len = inputs[Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH]
        cabin_len = inputs["cabin_len"]

        J[Aircraft.Fuselage.LENGTH, Aircraft.Fuselage.NOSE_FINENESS] = nose_height
        J[Aircraft.Fuselage.LENGTH, "nose_height"] = LoverD_nose
        J[Aircraft.Fuselage.LENGTH, Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH] = 1
        J[Aircraft.Fuselage.LENGTH, "cabin_len"] = 1
        J[Aircraft.Fuselage.LENGTH, Aircraft.Fuselage.TAIL_FINENESS] = cabin_height
        J[Aircraft.Fuselage.LENGTH, "cabin_height"] = LoverD_tail

        if self.options["aviary_options"].get_val(Aircraft.Fuselage.PROVIDE_SURFACE_AREA, units='unitless'):
            J[Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.WETTED_AREA_FACTOR] = 1

        else:
            J[Aircraft.Fuselage.WETTED_AREA, "cabin_height"] = fus_SA_factor * (
                2.5 * (LoverD_nose * nose_height + cockpit_len)
                + 3.14 * cabin_len
                + 2.1 * LoverD_tail * cabin_height
                + cabin_height * 2.1 * LoverD_tail
            )
            J[Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.NOSE_FINENESS] = (
                fus_SA_factor * cabin_height * 2.5 * nose_height
            )
            J[Aircraft.Fuselage.WETTED_AREA, "nose_height"] = (
                fus_SA_factor * cabin_height * 2.5 * LoverD_nose
            )
            J[Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH] = (
                fus_SA_factor * cabin_height * 2.5
            )
            J[Aircraft.Fuselage.WETTED_AREA, "cabin_len"] = (
                fus_SA_factor * 3.14 * cabin_height
            )
            J[Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.TAIL_FINENESS] = (
                fus_SA_factor * 2.1 * cabin_height**2
            )
            J[Aircraft.Fuselage.WETTED_AREA, Aircraft.Fuselage.WETTED_AREA_FACTOR] = cabin_height * (
                2.5 * (LoverD_nose * nose_height + cockpit_len)
                + 3.14 * cabin_len
                + 2.1 * LoverD_tail * cabin_height
            )

        J[Aircraft.TailBoom.LENGTH, Aircraft.Fuselage.NOSE_FINENESS] = nose_height
        J[Aircraft.TailBoom.LENGTH, "nose_height"] = LoverD_nose
        J[Aircraft.TailBoom.LENGTH, Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH] = 1
        J[Aircraft.TailBoom.LENGTH, "cabin_len"] = 1
        J[Aircraft.TailBoom.LENGTH, Aircraft.Fuselage.TAIL_FINENESS] = cabin_height
        J[Aircraft.TailBoom.LENGTH, "cabin_height"] = LoverD_tail


class FuselageGroup(om.Group):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        aviary_options = self.options['aviary_options']

        # outputs from parameters that are used in size but not outside of this group
        connected_input_outputs = ["cabin_height", "cabin_len", "nose_height"]

        parameters = self.add_subsystem(
            "parameters",
            FuselageParameters(
                aviary_options=aviary_options,
            ),
            promotes_inputs=["aircraft:*"],
            promotes_outputs=["aircraft:*"] + connected_input_outputs,
        )

        size = self.add_subsystem(
            "size",
            FuselageSize(aviary_options=aviary_options,),
            promotes_inputs=connected_input_outputs + ["aircraft:*"],
            promotes_outputs=["aircraft:*"],
        )

        self.set_input_defaults(
            Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, val=9.5, units="ft")



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class EngineSize(om.ExplicitComponent):
    """GASP engine geometry calculation."""

    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):
        add_aviary_input(self, Aircraft.Engine.REFERENCE_DIAMETER, 5.8)
        add_aviary_input(self, Aircraft.Engine.SCALE_FACTOR, 1.0)
        add_aviary_input(self, Aircraft.Nacelle.CORE_DIAMETER_RATIO, 1.25)
        add_aviary_input(self, Aircraft.Nacelle.FINENESS, 2)

        add_aviary_output(self, Aircraft.Nacelle.AVG_DIAMETER, val=0.0)
        add_aviary_output(self, Aircraft.Nacelle.AVG_LENGTH, val=0.0)
        add_aviary_output(self, Aircraft.Nacelle.SURFACE_AREA, val=0.0)

    def setup_partials(self):
        innames = [
            Aircraft.Engine.REFERENCE_DIAMETER,
            Aircraft.Engine.SCALE_FACTOR,
            Aircraft.Nacelle.CORE_DIAMETER_RATIO,
            Aircraft.Nacelle.FINENESS,
        ]
        self.declare_partials(Aircraft.Nacelle.AVG_DIAMETER, innames[:-1])
        self.declare_partials(Aircraft.Nacelle.AVG_LENGTH, innames)
        self.declare_partials(Aircraft.Nacelle.SURFACE_AREA, innames)

    def compute(self, inputs, outputs):
        d_ref = inputs[Aircraft.Engine.REFERENCE_DIAMETER]
        scale_fac = inputs[Aircraft.Engine.SCALE_FACTOR]
        d_nac_eng = inputs[Aircraft.Nacelle.CORE_DIAMETER_RATIO]
        ld_nac = inputs[Aircraft.Nacelle.FINENESS]

        d_eng = d_ref * np.sqrt(scale_fac)
        outputs[Aircraft.Nacelle.AVG_DIAMETER] = d_eng * d_nac_eng
        outputs[Aircraft.Nacelle.AVG_LENGTH] = (
            ld_nac * outputs[Aircraft.Nacelle.AVG_DIAMETER]
        )
        outputs[Aircraft.Nacelle.SURFACE_AREA] = (
            np.pi
            * outputs[Aircraft.Nacelle.AVG_DIAMETER]
            * outputs[Aircraft.Nacelle.AVG_LENGTH]
        )

    def compute_partials(self, inputs, J):
        d_ref = inputs[Aircraft.Engine.REFERENCE_DIAMETER]
        scale_fac = inputs[Aircraft.Engine.SCALE_FACTOR]
        d_nac_eng = inputs[Aircraft.Nacelle.CORE_DIAMETER_RATIO]
        ld_nac = inputs[Aircraft.Nacelle.FINENESS]

        tr = np.sqrt(scale_fac)
        d_eng = d_ref * tr
        d_nac = d_eng * d_nac_eng
        l_nac = d_nac * ld_nac

        J[Aircraft.Nacelle.AVG_DIAMETER, Aircraft.Engine.REFERENCE_DIAMETER] = tr * d_nac_eng
        J[Aircraft.Nacelle.AVG_DIAMETER, Aircraft.Engine.SCALE_FACTOR] = (
            d_nac_eng * d_ref / (2 * tr)
        )
        J[Aircraft.Nacelle.AVG_DIAMETER, Aircraft.Nacelle.CORE_DIAMETER_RATIO] = d_eng

        for wrt in [
            Aircraft.Engine.REFERENCE_DIAMETER,
            Aircraft.Engine.SCALE_FACTOR,
            Aircraft.Nacelle.CORE_DIAMETER_RATIO,
        ]:
            J[Aircraft.Nacelle.AVG_LENGTH, wrt] = (
                J[Aircraft.Nacelle.AVG_DIAMETER, wrt] * ld_nac
            )
            J[Aircraft.Nacelle.SURFACE_AREA, wrt] = np.pi * (
                J[Aircraft.Nacelle.AVG_DIAMETER, wrt] * l_nac
                + J[Aircraft.Nacelle.AVG_LENGTH, wrt] * d_nac
            )

        J[Aircraft.Nacelle.AVG_LENGTH, Aircraft.Nacelle.FINENESS] = d_nac
        J[Aircraft.Nacelle.SURFACE_AREA, Aircraft.Nacelle.FINENESS] = (
            np.pi * J[Aircraft.Nacelle.AVG_LENGTH, Aircraft.Nacelle.FINENESS] * d_nac
        )



'''
Dummy explicit component which serves as an input port for all variables in
the aircraft and mission hierarchy.
'''
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.core_promotes import core_mission_inputs


class VariablesIn(om.ExplicitComponent):
    '''
    Provides a central place to connect input variable information to a component
    but doesn't actually do anything on its own.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')
        self.options.declare(
            'meta_data', types=dict, default=_MetaData,
            desc='variable metadata associated with the variables to be passed through this port'
        )
        self.options.declare(
            'context', default='full', values=['full', 'mission'],
            desc='Limit to a subset of the aircraft and mission variables.'
        )

    def setup(self):
        aviary_options: AviaryValues = self.options['aviary_options']
        meta_data = self.options['meta_data']
        context = self.options['context']

        if context == 'mission':
            inputs = core_mission_inputs
        else:
            inputs = meta_data

        for key in inputs:
            # TODO temp line to ignore dynamic mission variables, will not work
            #      if names change to 'dynamic:mission:*'
            if ':' not in key:
                continue
            info = meta_data[key]

            if not info['option'] and ('aircraft:' in key or 'mission:' in key):
                # Since all the variable initial values are stored in aviary_options,
                # we can use the initial values to get the correct shape.
                val = info['default_value']
                if val is None:
                    val = 0.0
                item = val, info['units']
                val, units = aviary_options.get_item(key, item)
                if units == 'unitless' and info['units'] != 'unitless':
                    units = info['units']
                add_aviary_input(self, key, val=val, units=units, meta_data=meta_data)



"""
Run the a mission with a simple external component that computes the wing
and horizontal tail mass.
"""
from copy import deepcopy
import aviary.api as av

from aviary.examples.external_subsystems.simple_weight.simple_weight_builder import WingWeightBuilder


phase_info = deepcopy(av.default_height_energy_phase_info)
# Here we just add the simple weight system to only the pre-mission
phase_info['pre_mission']['external_subsystems'] = [WingWeightBuilder()]

prob = av.AviaryProblem()

# Load aircraft and options data from user
# Allow for user overrides here
prob.load_inputs('models/test_aircraft/aircraft_for_bench_FwFm.csv', phase_info)

prob.add_pre_mission_systems()

prob.add_phases()

prob.add_post_mission_systems()

# Link phases and variables
prob.link_phases()

prob.add_driver("SLSQP")

prob.add_design_variables()

prob.add_objective()

prob.setup()

prob.set_initial_guesses()

prob.run_aviary_problem(suppress_solver_print=True)

print('Engine Mass', prob.get_val(av.Aircraft.Engine.MASS))
print('Wing Mass', prob.get_val(av.Aircraft.Wing.MASS))
print('Horizontal Tail Mass', prob.get_val(av.Aircraft.HorizontalTail.MASS))

print('done')
